/*
** $Header: /u/spert/src/fltvec/nv/RCS/mat.c,v 1.14 1996/04/10 23:29:08 davidj Exp $
**
** Naive matrix routines
*/

#include <fltv-config.h>
#include <stddef.h>
#include <math.h>
#include <fltv-defs.h>
#include <fltv-scansrc.h>

/****************************************************************/
/* Accumlate a vector multiple times into a matrix */
/****************************************************************/

#ifdef DECL_nv_acc_vf_mf
void nv_acc_vf_mf(size_t mat_height, size_t vec_len, const float* vec,
		  float* mat);
#endif

#ifdef DEFN_nv_acc_vf_mf
void
nv_acc_vf_mf(size_t mat_height, size_t vec_len, const float* vec, float* mat)
{
    size_t i, j;

    const float* vec_ptr;
    float* mat_ptr = mat;

    for (i=0; i<mat_height; i++)
    {
	vec_ptr = vec;
	for (j=0; j<vec_len; j++)
	{
	    (*mat_ptr++) += (*vec_ptr++);
	}
    }
}
#endif /* #ifdef DEFN_nv_acc_vf_mf */

/****************************************************************/
/* Copy a vector multiple times into a matrix */
/****************************************************************/

#ifdef DECL_nv_copy_vf_mf
void nv_copy_vf_mf(size_t mat_height, size_t vec_len, const float* vec,
		   float* mat);
#endif

#ifdef DEFN_nv_copy_vf_mf
void
nv_copy_vf_mf(size_t mat_height, size_t vec_len, const float* vec, float* mat)
{
    size_t i, j;

    const float* vec_ptr;
    float* mat_ptr = mat;

    for (i=0; i<mat_height; i++)
    {
	vec_ptr = vec;
	for (j=0; j<vec_len; j++)
	{
	    (*mat_ptr++) = (*vec_ptr++);
	}
    }
}
#endif /* #ifdef DEFN_nv_copy_vf_mf */

/****************************************************************/
/* Copy an unstrided matrix to a strided matrix */
/****************************************************************/

#ifdef DECL_nv_copy_mf_smf
void nv_copy_mf_smf(size_t height, size_t in_width, size_t stride,
		    const float* from, float* to);
#endif

#ifdef DEFN_nv_copy_mf_smf
void
nv_copy_mf_smf(size_t height, size_t in_width, size_t stride,
	       const float* from, float* to)
{
    size_t i, j;

    for (i=0; i<height; i++)
    {
	for (j=0; j<in_width; j++)
	{
	    *(to+j) = *(from+j);
	}
	from += in_width;
	to += stride;
    }
}
#endif /* #ifdef DEFN_nv_copy_mf_smf */

/****************************************************************/
/* Copy a strided matrix to an unstrided matrix */
/****************************************************************/

#ifdef DECL_nv_copy_smf_mf
void nv_copy_smf_mf(size_t height, size_t width, size_t stride,
		    const float* from, float* to);
#endif

#ifdef DEFN_nv_copy_smf_mf
void
nv_copy_smf_mf(size_t height, size_t width, size_t stride,
	       const float* from, float* to)
{
    size_t i, j;

    for (i=0; i<height; i++)
    {
	for (j=0; j<width; j++)
	{
	    *(to+j) = *(from+j);
	}
	from += stride;
	to += width;
    }
}
#endif /* #ifdef DEFN_nv_copy_smf_mf */

/****************************************************************/
/* Matrix transpose */
/****************************************************************/

#ifdef DECL_nv_trans_mf_mf
void nv_trans_mf_mf(size_t in_height, size_t in_width, const float *in,
		    float *res);
#endif

#ifdef DEFN_nv_trans_mf_mf
void
nv_trans_mf_mf(size_t in_height, size_t in_width, const float *in, float *res)
{
    size_t i, j;
    const float *from;		/* Get elements from here */
    float *to;			/* Put elements here */
    float *tocol;		/* Current column in result */

    from = in;
    to = res;
    tocol = res;
    for (i=0; i<in_height; i++)
    {
	to = tocol;
	for (j=0; j<in_width; j++)
	{
	    *to = *from;
	    from++;
	    to += in_height;
	}
	tocol++;
    }
}
#endif /* #ifdef DEFN_nv_trans_mf_mf */


#ifdef DECL_nv_sumcol_mf_vf
/*
** Sum the columns of a matrix.
*/
void nv_sumcol_mf_vf(size_t rows, size_t cols, const float* mat,
		     float* res);
#endif

#ifdef DEFN_nv_sumcol_mf_vf
void
nv_sumcol_mf_vf(size_t rows, size_t cols, const float* mat, float* res)
{
    size_t i;
    size_t j;
    float* resp;		/* Pointer to result vector */

    resp = res;
    /* Initialize result to zero */
    for (j=0; j<cols; j++)
	*resp++ = 0.0f;

    /* Do the work, accessing the matrix sequentially */
    for (i=0; i<rows; i++)
    {
	resp = res;
	for(j=0; j<cols; j++)
	{
	    (*resp++) += (*mat++);
	}
    }
}
#endif

/****************************************************************/
/* Matrix-vector multiply, out_vec = in_matrix * in_vec. */
/****************************************************************/

#ifdef DECL_nv_mul_mfvf_vf
void nv_mul_mfvf_vf(size_t rows, size_t cols,
		    const float *in_mat, const float *in_vec, float *out_vec);
#endif

#ifdef DEFN_nv_mul_mfvf_vf
void
nv_mul_mfvf_vf(size_t rows, size_t cols,
               const float *in_mat, const float *in_vec, float *out_vec)
{

    size_t i;
    size_t j;

    for (i=0; i<rows; i++)
    {
        float sum = 0.0;
        const float *inp = in_vec;
        for (j=0; j<cols; j++)
        {
            sum += (*in_mat++) * (*inp++);
        }
        *out_vec++ = sum;
    }
}
#endif /* #ifdef DEFN_nv_mul_mfvf_vf */

/****************************************************************/
/* Vector-transpose-matrix multiply, out_vec = in_vec t * in_matrix. */
/****************************************************************/

#ifdef DECL_nv_mul_vfmf_vf
void nv_mul_vfmf_vf(size_t rows, size_t cols,
		    const float *in_vec, const float *in_mat, float *out_vec);
#endif

#ifdef DEFN_nv_mul_vfmf_vf
void
nv_mul_vfmf_vf(size_t rows, size_t cols,
            const float *in_vec, const float *in_mat, float *out_vec)
{
    size_t i;
    size_t j;

    for (i=0; i<cols; i++)
    {
        float sum = 0.0;
        const float *vp = in_vec;
        const float *mp = in_mat + i;
        for (j=0; j<rows; j++)
        {
            sum += (*mp) * (*vp++);
            mp += cols;
        }
        *out_vec++ = sum;
    }
}

#endif /* #ifdef DEFN_mul_vfmvf_vf */


/****************************************************************/
/* Matrix-matrix multiply, C = A*B */
/****************************************************************/

#ifdef DECL_nv_mul_mfmf_mf
void nv_mul_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
		    const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_mul_mfmf_mf
void
nv_mul_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
		  const float *A,const float *B,float *C)
{
  size_t i,j,k;
  for (i=0;i<Sm;i++)
  {
      for (j=0;j<Sn;j++)
      {
	  C[i*Sn+j] = 0.0f;
	  for (k=0;k<Sk;k++)
	      C[i*Sn+j] += A[i*Sk+k]*B[k*Sn+j];
      }
  }
}
#endif /* #ifdef DEFN_nv_mul_mfmf_mf */

/****************************************************************/
/* Matrix-matrix multiply/accumulate, C += A*B */
/****************************************************************/


#ifdef DECL_nv_mulacc_mfmf_mf
void nv_mulacc_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
		       const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_mulacc_mfmf_mf
void
nv_mulacc_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
		  const float *A,const float *B,float *C)
{
  size_t i,j,k;
  for (i=0;i<Sm;i++)
    for (j=0;j<Sn;j++)
      for (k=0;k<Sk;k++)
        C[i*Sn+j] += A[i*Sk+k]*B[k*Sn+j];
}
#endif /* #ifdef DEFN_nv_mulacc_mfmf_mf */


/****************************************************************/
/* Matrix-matrix multiply/accumulate, first input transposed, C += A^t*B */
/****************************************************************/

#ifdef DECL_nv_multnacc_mfmf_mf
void nv_multnacc_mfmf_mf(size_t Sk,size_t Sm,size_t Sn,
			 const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_multnacc_mfmf_mf
void
nv_multnacc_mfmf_mf(size_t Sk,size_t Sm,size_t Sn,
		    const float *A,const float *B,float *C)
{
  size_t i,j,k;
  for (i=0;i<Sm;i++)
    for (j=0;j<Sn;j++)
      for (k=0;k<Sk;k++)
        C[i*Sn+j] += A[k*Sm+i]*B[k*Sn+j];
}
#endif /* #ifdef DEFN_nv_multnacc_mfmf_mf */

/****************************************************************/
/* Matrix-matrix multiply/accumulate, second input transposed, C += A*B^t */
/****************************************************************/

#ifdef DECL_nv_mulntacc_mfmf_mf
void nv_mulntacc_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
			 const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_mulntacc_mfmf_mf
void
nv_mulntacc_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
		    const float *A,const float *B,float *C)
{
  size_t i,j,k;
  for (i=0;i<Sm;i++)
    for (j=0;j<Sn;j++)
      for (k=0;k<Sk;k++)
        C[i*Sn+j] += A[i*Sk+k]*B[j*Sk+k];
}
#endif /* #ifdef DEFN_nv_mulntacc_mfmf_mf */

/****************************************************************/
/* Matrix-matrix multiply/accumulate, both inputs transposed, C += A^t*B^t */
/****************************************************************/

#ifdef DECL_nv_multtacc_mfmf_mf
void nv_multtacc_mfmf_mf(size_t Sk,size_t Sm,size_t Sn,
			 const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_multtacc_mfmf_mf
void
nv_multtacc_mfmf_mf(size_t Sk,size_t Sm,size_t Sn,
		    const float *A,const float *B,float *C)
{
  size_t i,j,k;
  for (i=0;i<Sm;i++)
    for (j=0;j<Sn;j++)
      for (k=0;k<Sk;k++)
        C[i*Sn+j] += A[k*Sm+i]*B[j*Sk+k];
}
#endif /* #ifdef DEFN_nv_multtacc_mfmf_mf */


/****************************************************************/
/* Matrix-matrix scaled multiply/accumulate, C += A*B*s */
/****************************************************************/


#ifdef DECL_nv_mulacc_fmfmf_mf
void nv_mulacc_fmfmf_mf(size_t Sm,size_t Sk,size_t Sn, float scale,
			const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_mulacc_fmfmf_mf
void
nv_mulacc_fmfmf_mf(size_t Sm,size_t Sk,size_t Sn, float scale,
		   const float *A,const float *B,float *C)
{
    size_t i,j,k;
    float acc;
    
    for (i=0;i<Sm;i++)
    {
	for (j=0;j<Sn;j++)
	{
	    acc = 0.0f;
	    for (k=0;k<Sk;k++)
		acc += A[i*Sk+k]*B[k*Sn+j];
	    C[i*Sn+j] += acc * scale;
	}
    }
}
#endif /* #ifdef DEFN_nv_mulacc_fmfmf_mf */


/****************************************************************/
/* Matrix-matrix scaled multiply/accumulate,                    */
/* first input transposed, C += A^t*B*s                           */
/****************************************************************/

#ifdef DECL_nv_multnacc_fmfmf_mf
void nv_multnacc_fmfmf_mf(size_t Sk,size_t Sm,size_t Sn, float scale,
			 const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_multnacc_fmfmf_mf
void
nv_multnacc_fmfmf_mf(size_t Sk,size_t Sm,size_t Sn, float scale,
		    const float *A,const float *B,float *C)
{
    size_t i,j,k;
    float acc;

    for (i=0;i<Sm;i++)
    {
	for (j=0;j<Sn;j++)
	{
	    acc = 0.0f;
	    for (k=0;k<Sk;k++)
		acc += A[k*Sm+i]*B[k*Sn+j];
	    C[i*Sn+j] += acc * scale;
	}
    }
}
#endif /* #ifdef DEFN_nv_multnacc_fmfmf_mf */

/****************************************************************/
/* Matrix-matrix scaled multiply/accumulate                     */
/* second input transposed, C += A*B^t*s                        */
/****************************************************************/

#ifdef DECL_nv_mulntacc_fmfmf_mf
void nv_mulntacc_fmfmf_mf(size_t Sm,size_t Sk,size_t Sn, float scale,
			 const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_mulntacc_fmfmf_mf
void
nv_mulntacc_fmfmf_mf(size_t Sm,size_t Sk,size_t Sn, float scale,
		     const float *A,const float *B,float *C)
{
    size_t i,j,k;
    float acc;

    for (i=0;i<Sm;i++)
    {
	for (j=0;j<Sn;j++)
	{
	    acc = 0.0f;
	    for (k=0;k<Sk;k++)
		acc += A[i*Sk+k]*B[j*Sk+k];
	    C[i*Sn+j] += acc * scale;
	}
    }
}
#endif /* #ifdef DEFN_nv_mulntacc_mfmf_mf */

/****************************************************************/
/* Matrix-matrix scaled multiply/accumulate                     */
/* both inputs transposed, C += A^t*B^t*s                       */
/****************************************************************/

#ifdef DECL_nv_multtacc_fmfmf_mf
void nv_multtacc_fmfmf_mf(size_t Sk,size_t Sm,size_t Sn, float scale,
			 const float *A,const float *B,float *C);
#endif

#ifdef DEFN_nv_multtacc_fmfmf_mf
void
nv_multtacc_fmfmf_mf(size_t Sk,size_t Sm,size_t Sn, float scale,
		     const float *A,const float *B,float *C)
{
    size_t i,j,k;
    float acc;

    for (i=0;i<Sm;i++)
    {
	for (j=0;j<Sn;j++)
	{
	    acc = 0.0f;
	    for (k=0;k<Sk;k++)
		acc += A[k*Sm+i]*B[j*Sk+k];
	    C[i*Sn+j] += acc * scale;
	}
    }
}
#endif /* #ifdef DEFN_nv_multtacc_fmfmf_mf */


/****************************************************************/
/* Accumulate scaled outer-product in matrix. */
/****************************************************************/

#ifdef DECL_nv_mulacc_fvfvf_mf
void nv_mulacc_fvfvf_mf(size_t rows, size_t cols, float scale,
			const float *in_vec, const float *in_vec_t,
			float *res_mat);
#endif

#ifdef DEFN_nv_mulacc_fvfvf_mf
void
nv_mulacc_fvfvf_mf(size_t rows, size_t cols, float scale,
		   const float *in_vec, const float *in_vec_t, float *res_mat)
{
    size_t i;
    size_t j;

    for (i=0; i<rows; i++)
    {
        const float row_val = in_vec[i] * scale;

        const float *vtp = in_vec_t;

        for (j=0; j<cols; j++)
        {
            *res_mat++ += vtp[j] * row_val;
        }
    }
}
#endif /* #ifdef DEFN_nv_mulacc_dvdvd_md */

/****************************************************************/

#ifdef DEFN_nv_mulacc_dvdvd_md
void
nv_mulacc_dvdvd_md(size_t rows, size_t cols, double scale,
		   const double *in_vec, const double *in_vec_t,
		   double *res_mat)
{
    size_t i;
    size_t j;

    for (i=0; i<rows; i++)
    {
        const double row_val = in_vec[i] * scale;

        const double *vtp = in_vec_t;

        for (j=0; j<cols; j++)
        {
            *res_mat++ += vtp[j] * row_val;
        }
    }
}
#endif /* #ifdef DEFN_nv_mulacc_fvfvf_mf */
