/*
** $Header: /u/spert/src/fltvec/nv/RCS/vecele.c,v 1.21 1996/08/21 21:53:31 davidj Exp $
**
** Element wise vector operations - the quick and easy way
*/

#include <fltv-config.h>
#include <stddef.h>
#include <math.h>
#include <fltv-defs.h>
#include <fltv-scansrc.h>

/***********************************/
/*** one input vector/scalar ops ***/
/***********************************/

#ifdef DECL_nv_copy_f_vf
void nv_copy_f_vf(size_t len, float from, float* to);
#endif

#ifdef DEFN_nv_copy_f_vf
void
nv_copy_f_vf(size_t len, float from, float* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = from;
}
#endif

#ifdef DECL_nv_copy_d_vd
void nv_copy_d_vd(size_t len, double from, double* to);
#endif

#ifdef DEFN_nv_copy_d_vd
void
nv_copy_d_vd(size_t len, double from, double* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = from;
}
#endif

/***********************************/
/*** One input vector/vector ops ***/
/***********************************/

#ifdef DECL_nv_copy_vf_vf
void nv_copy_vf_vf(size_t len, const float* from, float* to);
#endif

#ifdef DEFN_nv_copy_vf_vf
void
nv_copy_vf_vf(size_t len, const float* from, float* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = *from++;
}
#endif

#ifdef DECL_nv_copy_vd_vd
void nv_copy_vd_vd(size_t len, const double* from, double* to);
#endif

#ifdef DEFN_nv_copy_vd_vd
void
nv_copy_vd_vd(size_t len, const double* from, double* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = *from++;
}
#endif

#ifdef DECL_nv_conv_vi32_vf
/*
** Convert a vector of 32 bit integers to single precision floating point,
** rounding if necessary.
*/
void nv_conv_vi32_vf(size_t len, const fltv_int32_t* in, float* res);
#endif

#ifdef DEFN_nv_conv_vi32_vf
void
nv_conv_vi32_vf(size_t len, const fltv_int32_t* in, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*res++ = (float) (*in++);
    }
}
#endif

#ifdef DECL_nv_abs_vf_vf
void nv_abs_vf_vf(size_t len, const float* from, float* to);
#endif

#ifdef DEFN_nv_abs_vf_vf
void nv_abs_vf_vf(size_t len, const float* from, float* to)
{
    size_t i;

    for (i=0; i<len; i++)
    {
#ifdef FLTV_HAVE_FABSF
	*to++ = fabsf(*from++);
#else
	*to++ = fabs(*from++);
#endif
    }
}
#endif

#ifdef DECL_nv_neg_vf_vf
void nv_neg_vf_vf(size_t len, const float* from, float* to);
#endif

#ifdef DEFN_nv_neg_vf_vf
void nv_neg_vf_vf(size_t len, const float* from, float* to)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*to++ = - (*from++);
    }
}
#endif

#ifdef DECL_nv_acc_vf_vf
void nv_acc_vf_vf(size_t n, const float *in_vec, float *res_vec);
#endif

#ifdef DEFN_nv_acc_vf_vf
void
nv_acc_vf_vf(size_t n, const float *in_vec, float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] += in_vec[i];
}
#endif

#ifdef DECL_nv_acc_vd_vd
void nv_acc_vd_vd(size_t n, const double *in_vec, double *res_vec);
#endif

#ifdef DEFN_nv_acc_vd_vd
void
nv_acc_vd_vd(size_t n, const double *in_vec, double *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] += in_vec[i];
}
#endif

/*
** Square root
*/

#ifdef DECL_nv_sqrt_vf_vf
void nv_sqrt_vf_vf(size_t len, const float* a, float* res);
#endif

#ifdef DEFN_nv_sqrt_vf_vf
void
nv_sqrt_vf_vf(size_t len, const float* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
#ifdef FLTV_HAVE_SQRTF
	*res++ = sqrtf(*a++);
#else
	*res++ = sqrt(*a++);
#endif
    }
}
#endif

/*
** Square - res = a * a
*/

#ifdef DECL_nv_sqr_vf_vf
void nv_sqr_vf_vf(size_t len, const float* a, float* res);
#endif

#ifdef DEFN_nv_sqr_vf_vf
void
nv_sqr_vf_vf(size_t len, const float* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*res++ = (*a) * (*a);
	a++;
    }
}
#endif

#ifdef DECL_nv_sqr_vd_vd
void nv_sqr_vd_vd(size_t len, const double* a, double* res);
#endif

#ifdef DEFN_nv_sqr_vd_vd
void
nv_sqr_vd_vd(size_t len, const double* a, double* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*res++ = (*a) * (*a);
	a++;
    }
}
#endif

/*
** Reciprocal - res = 1.0/a
*/

#ifdef DECL_nv_recip_vf_vf
void nv_recip_vf_vf(size_t len, const float* a, float* res);
#endif

#ifdef DEFN_nv_recip_vf_vf
void
nv_recip_vf_vf(size_t len, const float* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*res++ = 1.0f/(*a++);
    }
}
#endif

/*
** Vector sigmoid function = 1/(1+e^-x)
*/

#ifdef DECL_nv_sigmoid_vf_vf
void nv_sigmoid_vf_vf(size_t n, const float* in_vec, float* out_vec);
#endif

#ifdef DEFN_nv_sigmoid_vf_vf
void
nv_sigmoid_vf_vf(size_t n, const float* in_vec, float* out_vec)
{
    size_t i;

    for (i=0; i<n; i++)
    {
	out_vec[i] = sigmoid_f_f(in_vec[i]);
    }
}
#endif /* #ifdef DEFN_nv_sigmoid_vf_vf */


/*
** Vector derivative of sigmoid function.
** res = in * (1 - in)
*/

#ifdef DECL_nv_dsigmoid_vf_vf
void nv_dsigmoid_vf_vf(size_t n, const float* in_vec, float* out_vec);
#endif

#ifdef DEFN_nv_dsigmoid_vf_vf
void
nv_dsigmoid_vf_vf(size_t n, const float* in_vec, float* out_vec)
{
    size_t i;

    for (i=0; i<n; i++)
    {
        const float y = in_vec[i];

	out_vec[i] = (1.0f - y) * y;
    }
}
#endif /* #ifdef DEFN_nv_dsigmoid_vf_vf */


/*
** Vector softmax function = e^x_i / \sum(e^x_i)
*/

#ifdef DECL_nv_softmax_vf_vf
void nv_softmax_vf_vf(size_t n, const float* in_vec, float* out_vec);
#endif

/* TODO - this could use other fltvec functions - would be faster */
#ifdef DEFN_nv_softmax_vf_vf
void
nv_softmax_vf_vf(size_t n, const float* in_vec, float* out_vec)
{
    float *tmp;			/* A temporary vector to store exponents */
    float sumexp = 0.0f;	/* Sum of exponents */
    float scale;		/* 1/sum of exponents */
    size_t i;

    for (i=0; i<n; i++)
    {
	float e;		/* Exponent of current value */

#ifdef FLTV_HAVE_EXPF
	e = expf(in_vec[i]);
#else
	e = exp(in_vec[i]);
#endif
	out_vec[i] = e;
	sumexp += e;
    }
    scale = 1.0f/sumexp;
    for (i=0; i<n; i++)
    {
	out_vec[i] = out_vec[i] * scale;
    }
}

#endif /* #ifdef DEFN_nv_softmax_vf_vf */

#ifdef DECL_nv_conv_vf_vd
void nv_conv_vf_vd(size_t len, const float* a, double* res);
#endif

#ifdef DEFN_nv_conv_vf_vd
void
nv_conv_vf_vd(size_t len, const float* a, double* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (double) *a++;
}
#endif

#ifdef DECL_nv_conv_vd_vf
void nv_conv_vd_vf(size_t len, const double* a, float* res);
#endif

#ifdef DEFN_nv_conv_vd_vf
void
nv_conv_vd_vf(size_t len, const double* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (float) *a++;
}
#endif

/***********************************/
/*** Two input vector/scalar ops ***/
/***********************************/


#ifdef DECL_nv_add_vff_vf
void nv_add_vff_vf(size_t len, const float* a, float b, float* res);
#endif

#ifdef DEFN_nv_add_vff_vf
void
nv_add_vff_vf(size_t len, const float* a, float b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) + b;
}
#endif

#ifdef DECL_nv_sub_vff_vf
void nv_sub_vff_vf(size_t len, const float* a, float b, float* res);
#endif

#ifdef DEFN_nv_sub_vff_vf
void
nv_sub_vff_vf(size_t len, const float* a, float b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) - b;
}
#endif

#ifdef DECL_nv_sub_fvf_vf
void nv_sub_fvf_vf(size_t len, float a, const float* b, float* res);
#endif

#ifdef DEFN_nv_sub_fvf_vf
void
nv_sub_fvf_vf(size_t len, float a, const float* b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = a - (*b++);
}
#endif

#ifdef DECL_nv_mul_vff_vf
void nv_mul_vff_vf(size_t len, const float* a, float b, float* res);
#endif

#ifdef DEFN_nv_mul_vff_vf
void
nv_mul_vff_vf(size_t len, const float* a, float b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) * b;
}
#endif

#ifdef DECL_nv_mul_vdd_vd
void nv_mul_vdd_vd(size_t len, const double* a, double b, double* res);
#endif

#ifdef DEFN_nv_mul_vdd_vd
void
nv_mul_vdd_vd(size_t len, const double* a, double b, double* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) * b;
}
#endif

#ifdef DECL_nv_div_vff_vf
void nv_div_vff_vf(size_t len, const float* a, float b, float* res);
#endif

#ifdef DEFN_nv_div_vff_vf
void
nv_div_vff_vf(size_t len, const float* a, float b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) / b;
}
#endif

#ifdef DECL_nv_div_fvf_vf
void nv_div_fvf_vf(size_t len, float a, const float* b, float* res);
#endif

#ifdef DEFN_nv_div_fvf_vf
void
nv_div_fvf_vf(size_t len, float a, const float* b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = a / (*b++);
}
#endif

/***********************************/
/*** Three input vector/scalar ops ***/
/***********************************/

#ifdef DECL_nv_muladd_vffvf_vf
void nv_muladd_vffvf_vf(const size_t n, const float* xvec, const float s,
			const float *yvec, float* res);
#endif

#ifdef DECL_nv_muladd_vddvd_vd
void nv_muladd_vddvd_vd(const size_t n, const double* xvec, const double s,
			const double *yvec, double* res);
#endif

#ifdef DEFN_nv_muladd_vffvf_vf
#define FP_TYPE float
#define NV_MULADD_VFFVF_VF nv_muladd_vffvf_vf
#endif

#ifdef DEFN_nv_muladd_vddvd_vd
#define FP_TYPE double
#define NV_MULADD_VFFVF_VF nv_muladd_vddvd_vd
#endif

#if defined(DEFN_nv_muladd_vddvd_vd) || defined(DEFN_nv_muladd_vffvf_vf)

void
NV_MULADD_VFFVF_VF(const size_t n,
	  const FP_TYPE* xvec, const FP_TYPE s, const FP_TYPE *yvec,
	  FP_TYPE* res)
{
    size_t i;
    for (i=0; i<n; i++)
        res[i] = s * xvec[i] + yvec[i];
}

#undef FP_TYPE
#endif


/***********************************/
/*** Two input vector/vector ops ***/
/***********************************/


#ifdef DECL_nv_add_vfvf_vf
void nv_add_vfvf_vf(size_t n, const float *in_vec1, const float *in_vec2,
		    float *res_vec);
#endif

#ifdef DEFN_nv_add_vfvf_vf
void
nv_add_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] + in_vec2[i];
}
#endif

#ifdef DECL_nv_add_vdvd_vd
void nv_add_vdvd_vd(size_t n, const double *in_vec1, const double *in_vec2,
		    double *res_vec);
#endif

#ifdef DEFN_nv_add_vdvd_vd
void
nv_add_vdvd_vd(size_t n,
	       const double *in_vec1, const double *in_vec2,
	       double *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] + in_vec2[i];
}
#endif


#ifdef DECL_nv_sub_vfvf_vf
void nv_sub_vfvf_vf(size_t n, const float *in_vec1, const float *in_vec2,
		    float *res_vec);
#endif

#ifdef DEFN_nv_sub_vfvf_vf
void
nv_sub_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] - in_vec2[i];
}
#endif

#ifdef DECL_nv_sub_vdvd_vd
void nv_sub_vdvd_vd(size_t n, const double *in_vec1, const double *in_vec2,
		    double *res_vec);
#endif

#ifdef DEFN_nv_sub_vdvd_vd
void
nv_sub_vdvd_vd(size_t n,
	       const double *in_vec1, const double *in_vec2,
	       double *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] - in_vec2[i];
}
#endif

#ifdef DECL_nv_mul_vfvf_vf
void nv_mul_vfvf_vf(size_t n, const float *in_vec1, const float *in_vec2,
		    float *res_vec);
#endif

#ifdef DEFN_nv_mul_vfvf_vf
void
nv_mul_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] * in_vec2[i];
}
#endif

#ifdef DECL_nv_mul_vdvd_vd
void nv_mul_vdvd_vd(size_t n, const double *in_vec1, const double *in_vec2,
		    double *res_vec);
#endif

#ifdef DEFN_nv_mul_vdvd_vd
void
nv_mul_vdvd_vd(size_t n,
	       const double *in_vec1, const double *in_vec2,
	       double *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] * in_vec2[i];
}
#endif

#ifdef DECL_nv_div_vfvf_vf
void nv_div_vfvf_vf(size_t n, const float *in_vec1, const float *in_vec2,
		    float *res_vec);
#endif

#ifdef DEFN_nv_div_vfvf_vf
void
nv_div_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] / in_vec2[i];
}
#endif

/* A vector/scalar accumulating multiply */

#ifdef DECL_nv_mulacc_vff_vf
void nv_mulacc_vff_vf(size_t n, const float *in, float scale, float *acc);
#endif

#ifdef DEFN_nv_mulacc_vff_vf
void
nv_mulacc_vff_vf(size_t n, const float *in, float scale, float *acc)
{
    size_t i;
    for (i=0; i<n; i++)
        acc[i] += scale * in[i];
}
#endif


/*******************/
/*** Wierdo ops! ***/
/*******************/

/*
** Multiply second vector by derivative of sigmoid by first vector.
** res = in1 * (1 - in1) * in2
*/

#ifdef DECL_nv_dsigmoid_mul_vf_vf_vf
void nv_dsigmoid_mul_vf_vf_vf(size_t n,
			      const float *in_vec1, const float *in_vec2,
			      float *res_vec);
#endif

#ifdef DEFN_nv_dsigmoid_mul_vf_vf_vf
void
nv_dsigmoid_mul_vf_vf_vf(size_t n,
			 const float *in_vec1, const float *in_vec2,
			 float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
    {
        const float y = in_vec1[i];

        res_vec[i] = (1.0f - y) * y * (in_vec2[i]);
    }
}
#endif

